(* tools for arithmetic
   WN.8.3.01
   use"../knowledge/Atools.ML";
   use"knowledge/Atools.ML";
   use"Atools.ML";
   *)

(*
copy from doc/math-eng.tex WN.28.3.03

\section{Coding standards}
\subsection{Rule sets}
The actual version of the coding standards for rulesets is in {\tt /Isa02/Tools.ML} where it can be viewed using the knowledge browsers.

There are rulesets visible to the student, and there are rulesets visible (in general) only for math authors. There are also rulesets which {\em must} exist for {\em each} theory; these contain the identifier of the respective theory (including all capital letters) as indicated by {\it Thy} below.
\begin{description}

\item [norm\_{\it Thy}] exists for each theory, and {\em efficiently} calculates a normalform for all terms which can be expressed by the definitions of the respective theory (and the respective parents).

\item [simplify\_{\it Thy}] exists for each theory, and calculates a normalform for all terms which can be expressed by the definitions of the respective theory (and the respective parents) such, that the rewrites can be presented to the student.

\item [calculate\_{\it Thy}] exists for each theory, and evaluates terms with numerical constants only (i.e. all terms which can be expressed by the definitions of the respective theory and the respective parent theories). In particular, this ruleset includes evaluating in/equalities with numerical constants only.

\end{description}
The following rulesets are used for internal purposes and usually invisible to the (naive) user:
\begin{description}

\item [*\_erls] 
\item [*\_prls] 
\item [*\_srls] 

\end{description}
{\tt append_rls, merge_rls, remove_rls}
*)

"******* Atools.ML begin *******";
theory' := overwritel (!theory', [("Atools.thy",Atools.thy)]);

(** evaluation of numerals and special predicates on the meta-level **)
(*-------------------------functions---------------------*)
local (* rlang 09.02 *)
    (*.a 'c is coefficient of v' if v does occur in c.*)
    fun coeff_in v c = v mem (vars c);
in
    fun occurs_in v t = coeff_in v t;
end;

(*("occurs_in", ("Atools.occurs'_in", eval_occurs'_in ""))*)
fun eval_occurs_in _ _ 
	     (p as (Const ("Atools.occurs'_in",_) $ v $ t)) _ =
    ((*writeln("@@@ eval_occurs_in: v= "^(term2str v));
     writeln("@@@ eval_occurs_in: t= "^(term2str t));*)
     if occurs_in v t
    then Some ((term2str p) ^ " = True",
	  Trueprop $ (mk_equality (p, HOLogic.true_const)))
    else Some ((term2str p) ^ " = True",
	  Trueprop $ (mk_equality (p, HOLogic.false_const))))
  | eval_occurs_in _ _ _ _ = None;
   
(*evaluate 'is_atom'*)
(*("is_atom",("Atools.is'_atom",eval_is_atom "#is_atom_"))*)
fun eval_is_atom (thmid:string) _ (t as (Const(op0,_) $ arg)) thy = 
    (case arg of 
	 Free (n,_) => Some (mk_thmid thmid op0 n "", 
			      Trueprop $ (mk_equality (t, true_as_term)))
       | _ => Some (mk_thmid thmid op0 "" "", 
		    Trueprop $ (mk_equality (t, false_as_term))))
  | eval_is_atom _ _ _ _ = None;

(*evaluate 'is_even'*)
fun even i = (i div 2) * 2 = i;
(*("is_even",("Atools.is'_even",eval_is_even "#is_even_"))*)
fun eval_is_even (thmid:string) _ (t as (Const(op0,_) $ arg)) thy = 
    (case arg of 
	Free (n,_) =>
	 (case int_of_str n of
	      Some i =>
	      if even i then Some (mk_thmid thmid op0 n "", 
				   Trueprop $ (mk_equality (t, true_as_term)))
	      else Some (mk_thmid thmid op0 "" "", 
			 Trueprop $ (mk_equality (t, false_as_term)))
	    | _ => None)
       | _ => None)
  | eval_is_even _ _ _ _ = None; 

(*evaluate 'is_const'*)
(*("is_const",("Atools.is'_const",eval_const "#is_const_"))*)
fun eval_const (thmid:string) _ (t as (Const(op0,t0) $ arg)) (thy:theory) = 
    (case arg of 
       Const (n1,_) =>
	 Some (mk_thmid thmid op0 n1 "", 
	       Trueprop $ (mk_equality (t, false_as_term)))
     | Free (n1,_) =>
	 if is_numeral n1
	   then Some (mk_thmid thmid op0 n1 "", 
		      Trueprop $ (mk_equality (t, true_as_term)))
	 else Some (mk_thmid thmid op0 n1 "", 
		    Trueprop $ (mk_equality (t, false_as_term)))
     | _ => None)
  | eval_const _ _ _ _ = None; 

(*. evaluate binary, associative, commutative operators: *,+,^ .*)
(*("plus"    ,("op +"        ,eval_binop "#add_")),
  ("times"   ,("op *"        ,eval_binop "#mult_")),
  ("power_"  ,("Atools.pow"  ,eval_binop "#power_"))*)

(* val (thmid,op_,t as(Const (op0,t0) $ Free (n1,t1) $ Free(n2,t2)),thy) =
       ("xxxxxx",op_,t,thy);
   *)
fun mk_thmid_f thmid ((v11, v12), (p11, p12)) ((v21, v22), (p21, p22))  = 
    thmid ^ "Float ((" ^ 
    (string_of_int v11)^","^(string_of_int v12)^"), ("^
    (string_of_int p11)^","^(string_of_int p12)^")) __ (("^
    (string_of_int v21)^","^(string_of_int v22)^"), ("^
    (string_of_int p21)^","^(string_of_int p22)^"))";

fun eval_binop (thmid:string) (op_:string) 
	       (t as ( Const(op0,t0) $ 
			    (Const(op0',t0') $ v $ t1) $ t2)) 
	       thy =                                     (*binary . (v.n1).n2*)
    if op0 = op0' then
	case (numeral t1, numeral t2) of
	    (Some n1, Some n2) =>
	    let val (T1,T2,Trange) = dest_binop_typ t0
		val res = calc op0 n1 n2
		val rhs = var_op_float v op_ t0 T1 res
		val prop = Trueprop $ (mk_equality (t, rhs))
	    in Some (mk_thmid_f thmid n1 n2, prop) end
	  | _ => None
    else None
  | eval_binop (thmid:string) (op_:string) 
	       (t as 
		  (Const (op0, t0) $ t1 $ 
			 (Const (op0', t0') $ t2 $ v))) 
	       thy =                                     (*binary . n1.(n2.v)*)
  if op0 = op0' then
	case (numeral t1, numeral t2) of
	    (Some n1, Some n2) =>
	    let val (T1,T2,Trange) = dest_binop_typ t0
		val res = calc op0 n1 n2
		val rhs = float_op_var v op_ t0 T1 res
		val prop = Trueprop $ (mk_equality (t, rhs))
	    in Some (mk_thmid_f thmid n1 n2, prop) end
	  | _ => None
  else None
    
  | eval_binop (thmid:string) (op_:string)
	       (t as (Const (op0,t0) $ t1 $ t2)) thy =       (*binary . n1.n2*)
    (case (numeral t1, numeral t2) of
	 (Some n1, Some n2) =>
	 let val (T1,T2,Trange) = dest_binop_typ t0;
	     val res = calc op0 n1 n2;
	     val rhs = term_of_float Trange res;
	     val prop = Trueprop $ (mk_equality (t, rhs));
	 in Some (mk_thmid_f thmid n1 n2, prop) end
       | _ => None)
  | eval_binop _ _ _ _ = None; 
(*
> val Some (thmid, t) = eval_binop "#add_" "op +" (str2term "-1 + 2") thy;
> term2str t;
val it = "-1 + 2 = 1"
> val t = str2term "-1 * (-1 * a)";
> val Some (thmid, t) = eval_binop "#mult_" "op *" t thy;
> term2str t;
val it = "-1 * (-1 * a) = 1 * a"*)



(*.evaluate < and <= for numerals.*)
(*("le"      ,("op <"        ,eval_equ "#less_")),
  ("leq"     ,("op <="       ,eval_equ "#less_equal_"))*)
fun eval_equ (thmid:string) (op_:string) (t as 
	       (Const (op0,t0) $ Free (n1,t1) $ Free(n2,t2))) thy = 
    (case (int_of_str n1, int_of_str n2) of
	 (Some n1', Some n2') =>
  if calc_equ (strip_thy op0) (n1', n2')
    then Some (mk_thmid thmid op0 n1 n2, 
	  Trueprop $ (mk_equality (t, true_as_term)))
  else Some (mk_thmid thmid op0 n1 n2,  
	  Trueprop $ (mk_equality (t, false_as_term)))
       | _ => None)
    
  | eval_equ _ _ _ _ = None;


(*evaluate identity
> reflI;
val it = "(?t = ?t) = True"
> val t = str2term "x = 0";
> val None = rewrite_ thy dummy_ord e_rls false reflI t;

> val t = str2term "1 = 0";
> val None = rewrite_ thy dummy_ord e_rls false reflI t;
----------- thus needs Calc !
> val t = str2term "0 = 0";
> val Some (t',_) = rewrite_ thy dummy_ord e_rls false reflI t;
> term2str t';
val it = "True"

val t = str2term "Not (x = 0)";
atomt t; term2str t;
*** -------------
*** Const ( Not)
*** . Const ( op =)
*** . . Free ( x, )
*** . . Free ( 0, )
val it = "x ~= 0" : string*)

(*.evaluate identity on the term-level, =!= ,i.e. without evaluation of 
  the arguments: thus this Calculation MUST be at the head of an rls*)
(*("ident"   ,("Atools.ident",eval_ident "#ident_")):calc*)
fun eval_ident (thmid:string) (op_:string) (t as 
	       (Const (op0,t0) $ t1 $ t2 )) thy = 
  if t1 = t2
    then Some (mk_thmid thmid op0 
	       ("("^(Sign.string_of_term (sign_of thy) t1)^")")
	       ("("^(Sign.string_of_term (sign_of thy) t2)^")"), 
	  Trueprop $ (mk_equality (t, true_as_term)))
  else Some (mk_thmid thmid op0  
	       ("("^(Sign.string_of_term (sign_of thy) t1)^")")
	       ("("^(Sign.string_of_term (sign_of thy) t2)^")"),  
	  Trueprop $ (mk_equality (t, false_as_term)))
  | eval_ident _ _ _ _ = None;
(* TODO
> val t = str2term "x =!= 0";
> val Some (str, t') = eval_ident "ident_" "b" t thy;
> term2str t';
val str = "ident_(x)_(0)" : string
val it = "(x =!= 0) = False" : string                                
> val t = str2term "1 =!= 0";
> val Some (str, t') = eval_ident "ident_" "b" t thy;
> term2str t';
val str = "ident_(1)_(0)" : string 
val it = "(1 =!= 0) = False" : string                                       
> val t = str2term "0 =!= 0";
> val Some (str, t') = eval_ident "ident_" "b" t thy;
> term2str t';
val str = "ident_(0)_(0)" : string
val it = "(0 =!= 0) = True" : string
*)

(*.evaluate identity of terns, which stay ready for evaluation in turn;
  thus returns False only for numerals.*)
(*("equal"   ,("op =",eval_equal "#equal_")):calc*)
fun eval_equal (thmid:string) (op_:string) (t as 
	       (Const (op0,t0) $ t1 $ t2 )) thy = 
  if t1 = t2
    then Some (mk_thmid thmid op0 
	       ("("^(Sign.string_of_term (sign_of thy) t1)^")")
	       ("("^(Sign.string_of_term (sign_of thy) t2)^")"), 
	  Trueprop $ (mk_equality (t, true_as_term)))
  else (case (is_num t1, is_num t2) of
	   (true, true) => 
	   if free2int t1 = free2int t2
	   then Some (mk_thmid thmid op0 
			       ("("^(term2str t1)^")") ("("^(term2str t2)^")"),
		      Trueprop $ (mk_equality (t, true_as_term)))
	   else Some (mk_thmid thmid op0  
			       ("("^(term2str t1)^")") ("("^(term2str t2)^")"),
		      Trueprop $ (mk_equality (t, false_as_term)))
	 | _ => None)
  | eval_equal _ _ _ _ = None;
(*
> val t = str2term "(x + 1) = (x + 1)";
> val Some (str, t') = eval_equal "equal_" "b" t thy;
> term2str t';
val str = "equal_(x + 1)_(x + 1)" : string
val it = "(x + 1 = x + 1) = True" : string
> val t = str2term "x = 0";
> val None = eval_equal "equal_" "b" t thy;

> val t = str2term "1 = 0";
> val Some (str, t') = eval_equal "equal_" "b" t thy;
> term2str t';
val str = "equal_(1)_(0)" : string 
val it = "(1 = 0) = False" : string
> val t = str2term "0 = 0";
> val Some (str, t') = eval_equal "equal_" "b" t thy;
> term2str t';
val str = "equal_(0)_(0)" : string
val it = "(0 = 0) = True" : string
*)


(** evaluation on the metalevel **)

(*. evaluate HOL.divide .*)
(*("divide_" ,("HOL.divide"  ,eval_cancel "#divide_"))*)
fun eval_cancel (thmid:string) _ (t as 
	       (Const (op0,t0) $ Free (n1,t1) $ Free(n2,t2))) thy = 
    (case (int_of_str n1, int_of_str n2) of
	 (Some n1', Some n2') =>
  let 
    (*val _= writeln"@@@ eval_cancel Some";*)
    val sg = sign2 n1' n2';
    val (T1,T2,Trange) = dest_binop_typ t0;
    val gcd' = gcd (abs n1') (abs n2');
  in if gcd' = abs n2' 
     then let val rhs = term_of_num Trange (sg * (abs n1') div gcd')
	      val prop = Trueprop $ (mk_equality (t, rhs))
	  in Some (mk_thmid thmid op0 n1 n2, prop) end     
     else if 0 < n2' andalso gcd' = 1 then None
     else let val rhs = num_op_num T1 T2 (op0,t0) (sg * (abs n1') div gcd')
				   ((abs n2') div gcd')
	      val prop = Trueprop $ (mk_equality (t, rhs))
	  in Some (mk_thmid thmid op0 n1 n2, prop) end
  end
       | _ => ((*writeln"@@@ eval_cancel None";*)None))

  | eval_cancel _ _ _ _ = None;


local

open Term;

in
fun termlessI (_:subst) uv = termless uv;
fun term_ordI (_:subst) uv = term_ord uv;
end;


(** rule set, for evaluating list-expressions in scripts 8.01.02

val simprls = 
  Rls{preconds = [], rew_ord = ("tless_true",tless_true), 
      rules = [Calc ("Atools.ident",eval_ident "#ident_"),
	       Calc ("Atools.Var",eval_var "#Var_"),
	       Calc ("Atools.Length",eval_Length "#Length_"),
	       Calc ("Atools.Nth",eval_Nth "#Nth_"),
	       Calc ("op <",eval_equ "#less_"),
	       Calc ("op <=",eval_equ "#less_equal_"),
	       Calc ("Atools.ident",eval_ident "#ident_"),

	       Thm ("if_True",num_str if_True),
	       Thm ("if_False",num_str if_False)
	       ],
      scr = Script ((term_of o the o (parse thy)) 
      "empty_script")
      }:rls;      *)

(** rule set for evaluating listexpr in scripts FIXXXME -> ListG.ML**)
val list_rls = 
  Rls{id="list_rls",preconds = [], rew_ord = ("e_rew_ord",e_rew_ord), 
      erls = e_rls, srls = Erls, calc = [], asm_thm=[],
      rules = (*8.01: copied from*)
      [Thm ("length_Nil_",num_str length_Nil_),
       Thm ("length_Cons_",num_str length_Cons_),
(*       Thm ("",),
       Thm ("",),
       Thm ("",),
       Thm ("",),
*)
       Calc ("op *",eval_binop "#mult_"),
       Calc ("op +", eval_binop "#add_"), 

       Calc ("Atools.ident",eval_ident "#ident_"),
       Calc ("Atools.Var",eval_var "#Var_"),
       Calc ("Atools.Length",eval_Length "#Length_"),(*8.01 Thm length_Nil_ !*)
       Calc ("Atools.Nth",eval_Nth "#Nth_"),         (*8.01 Thm  nth_Cons_ !!*)
       Calc ("op <",eval_equ "#less_"),
       Calc ("op <=",eval_equ "#less_equal_"),
       Calc ("Atools.ident",eval_ident "#ident_"),
       
       Thm ("if_True",num_str if_True),
       Thm ("if_False",num_str if_False)
       ], scr = EmptyScr}:rls;

ruleset' := overwritel (!ruleset',
  [("list_rls",list_rls)
   ]);

val tless_true = dummy_ord;
rew_ord' := overwritel (!rew_ord',
			[("tless_true", tless_true)]);

val calculate_Atools = 
    append_rls "calculate_Atools" e_rls
               [Calc ("op <",eval_equ "#less_"),
		Calc ("op <=",eval_equ "#less_equal_"),
		Calc ("op =",eval_equal "#equal_"),

		Thm  ("real_unari_minus",num_str real_unari_minus),
		Calc ("op +",eval_binop "#add_"),
		Calc ("op -",eval_binop "#subtract_"),
		Calc ("op *",eval_binop "#mult_")
		];

val Atools_erls = 
    append_rls "Atools_erls" e_rls
               [Thm ("not_true",num_str not_true),
		Thm ("not_false",num_str not_false),
		Thm ("and_true",and_true),
		Thm ("and_false",and_false),
		Thm ("or_true",or_true),
		Thm ("or_false",or_false)		
		];

(*val atools_erls = waere zu testen ...
    merge_rls calculate_Atools
	      (append_rls Atools_erls (*i.A. zu viele rules*)
			  [Calc ("Atools.ident",eval_ident "#ident_"),    
			   Calc ("Atools.is'_const",eval_const "#is_const_"),
			   Calc ("Atools.occurs'_in",
				 eval_occurs_in "#occurs_in"),    
			   Calc ("Tools.matches",eval_matches "#matches")
			   ] (*i.A. zu viele rules*)
			  );*)
val atools_erls = 
  Rls {id="atools_erls",preconds = [], rew_ord = ("termlessI",termlessI), 
      erls = e_rls, srls = Erls, calc = [], asm_thm = [],
      rules = [Thm ("refl",num_str refl),
		Thm ("le_refl",num_str le_refl),
		Thm ("radd_left_cancel_le",num_str radd_left_cancel_le),
		Thm ("not_true",num_str not_true),
		Thm ("not_false",num_str not_false),
		Thm ("and_true",and_true),
		Thm ("and_false",and_false),
		Thm ("or_true",or_true),
		Thm ("or_false",or_false),
		Thm ("and_commute",num_str and_commute),
		Thm ("or_commute",num_str or_commute),
		
		Calc ("op <",eval_equ "#less_"),
		Calc ("op <=",eval_equ "#less_equal_"),
		
		Calc ("Atools.ident",eval_ident "#ident_"),    
		Calc ("Atools.is'_const",eval_const "#is_const_"),
		Calc ("Atools.occurs'_in",eval_occurs_in ""),    
		Calc ("Tools.matches",eval_matches "")
	       ],
      scr = Script ((term_of o the o (parse thy)) 
      "empty_script")
      }:rls;
ruleset' := overwritel 
		(!ruleset',
		 [("atools_erls",atools_erls)(*FIXXXME:del with rls.rls'*)
		  ]);
"******* Atools.ML end *******";


