(* tools for applications of differetiation
 use"DiffApp.ML";
 use"knowledge/DiffApp.ML";
 use"../knowledge/DiffApp.ML";
 *)


(** interface isabelle -- isac **)

theory' := overwritel (!theory', [("DiffApp.thy",DiffApp.thy)]);

val eval_rls =
  Rls {id="eval_rls",preconds = [], rew_ord = ("termlessI",termlessI), 
      erls = e_rls, srls = Erls, calc = [], asm_thm = [],
      rules = [Thm ("refl",num_str refl),
		Thm ("le_refl",num_str le_refl),
		Thm ("radd_left_cancel_le",num_str radd_left_cancel_le),
		Thm ("not_true",num_str not_true),
		Thm ("not_false",num_str not_false),
		Thm ("and_true",and_true),
		Thm ("and_false",and_false),
		Thm ("or_true",or_true),
		Thm ("or_false",or_false),
		Thm ("and_commute",num_str and_commute),
		Thm ("or_commute",num_str or_commute),
		
		Calc ("op <",eval_equ "#less_"),
		Calc ("op <=",eval_equ "#less_equal_"),
		
		Calc ("Atools.ident",eval_ident "#ident_"),    
		Calc ("Atools.is'_const",eval_const "#is_const_"),
		Calc ("Atools.occurs'_in",eval_occurs_in ""),    
		Calc ("Tools.matches",eval_matches "")
	       ],
      scr = Script ((term_of o the o (parse thy)) 
      "empty_script")
      }:rls;
ruleset' := overwritel 
		(!ruleset',
		 [("eval_rls",atools_erls)(*FIXXXME:del with rls.rls'*)
		  ]);


(** problem types **)

store_pbt
 (prep_pbt DiffApp.thy
 (["maximum_of","function"],
  [("#Given" ,["fixedValues fix_"]),
   ("#Find"  ,["maximum m_","valuesFor vs_"]),
   ("#Relate",["relations rs_"])
  ],
  e_rls, None, []));

store_pbt
 (prep_pbt DiffApp.thy
 (["make","function"]:pblID,
  [("#Given" ,["functionOf f_","boundVariable v_","equalities eqs_"]),
   ("#Find"  ,["functionTerm f_1_"])
  ],
  e_rls, None, []));

store_pbt
 (prep_pbt DiffApp.thy
 (["on_interval","maximum_of","function"]:pblID,
  [("#Given" ,["functionTerm t_","boundVariable v_","interval itv_"]),
   ("#Find"  ,["maxArgument v_0_"])
  ],
  e_rls, None, []));

store_pbt
 (prep_pbt DiffApp.thy
 (["tool"]:pblID,
  [],
  e_rls, None, []));

store_pbt
 (prep_pbt DiffApp.thy
 (["find_values","tool"]:pblID,
  [("#Given" ,["maxArgument ma_","functionTerm f_","boundVariable v_"]),
   ("#Find"  ,["valuesFor vls_"]),
   ("#Relate",["additionalRels rs_"])
  ],
  e_rls, None, []));


(** methods, scripts not yet implemented **)

methods:= overwritel (!methods,
[
 prep_met 
 (("DiffApp.thy","max_by_calculus"):metID,
  [("#Given" ,["fixedValues fix_","boundVariable v_","interval itv_",
	       "errorBound err_"]),
    ("#Find"  ,["maximum m_","valuesFor vs_"]),
    ("#Relate",["relations rs_"])
    ],
  {rew_ord'="tless_true",rls'=eval_rls,calc=[], srls = e_rls,prls=e_rls,
   asm_rls=[],asm_thm=[]},
(*---- see test-script.sml
  "Script Maximum_value(fix_::bool list)(m_::real) (rs_::bool list)\
   \      (v_::real) (itv_::real set) (err_::bool) =          \ 
   \ (let e_ = (hd o (filter (Testvar m_))) rs_;              \
   \      t_ = (if 1 < Length rs_                            \
   \           then (SubProblem (Reals_,[make,function],(Isac_,no_met))\
   \                     [real_ m_, real_ v_, bool_list_ rs_])\
   \           else (hd rs_));                                \
   \      (mx_::real) = SubProblem (Reals_,[on_interval,max_of,function], \
   \                                (Isac_,maximum_on_interval))\
   \                               [bool_ t_, real_ v_, real_set_ itv_]\
   \ in ((SubProblem (Reals_,[find_values,tool],(Isac_,find_values))   \
   \      [real_ mx_, real_ (Rhs t_), real_ v_, real_ m_,     \
   \       bool_list_ (dropWhile (ident e_) rs_)])::bool list))"*)
  "Script Maximum_value(fix_::bool list)(m_::real) (rs_::bool list)\
   \      (v_::real) (itv_::real set) (err_::bool) =           \ 
   \ (let (t_::bool) = ((SubProblem (Reals_,[make,function],(Isac_,no_met)) \
   \                     [real_ m_, real_ v_, bool_list_ rs_])::bool)\
   \ in (rs_ ::bool list))"
  )
,
 prep_met 
 (("DiffApp.thy","make_fun_by_new_variable"):metID,
   [("#Given" ,["functionOf f_","boundVariable v_","equalities eqs_"]),
    ("#Find"  ,["functionTerm f_1_"])
    ],
   {rew_ord'="tless_true",rls'=eval_rls,srls = list_rls,prls=e_rls,
    calc=[],asm_rls=[],asm_thm=[]},
  "Script Make_fun_by_new_variable (f_::real) (v_::real)     \
   \      (eqs_::bool list) =                                 \
   \(let h_ = (hd o (filter (Testvar m_))) eqs_;             \
   \     es_ = dropWhile (ident h_) eqs_;                    \
   \     vs_ = dropWhile (ident f_) (Var h_);                \
   \     v1_ = Nth 1 vs_;                                   \
   \     v2_ = Nth 2 vs_;                                   \
   \     e1_ = (hd o (filter (Testvar v1_))) es_;            \
   \     e2_ = (hd o (filter (Testvar v2_))) es_;            \
   \  (s_1::bool list) = (SubProblem (Reals_,[univar,equation],(Isac_,no_met))\
   \                    [bool_ e1_, real_ v1_]);\
   \  (s_2::bool list) = (SubProblem (Reals_,[univar,equation],(Isac_,no_met))\
   \                    [bool_ e2_, real_ v2_])\
   \in Substitute [(v_1, (Rhs o hd) s_1),(v_2, (Rhs o hd) s_2)] h_)"
)
,
 prep_met 
(("DiffApp.thy","make_fun_by_explicit"):metID,
   [("#Given" ,["functionOf f_","boundVariable v_","equalities eqs_"]),
    ("#Find"  ,["functionTerm f_1_"])
    ],
   {rew_ord'="tless_true",rls'=eval_rls,calc=[],srls=list_rls,prls=e_rls,
    asm_rls=[],asm_thm=[]},
   "Script Make_fun_by_explicit (f_::real) (v_::real)         \
   \      (eqs_::bool list) =                                 \
   \ (let h_ = (hd o (filter (Testvar m_))) eqs_;             \
   \      e1_ = hd (dropWhile (ident h_) eqs_);               \
   \      vs_ = dropWhile (ident f_) (Var h_);                \
   \      v1_ = hd (dropWhile (ident v_) vs_);                \
   \      (s1::bool list)=(SubProblem(Reals_,[univar,equation],(Isac_,no_met))\
   \                          [bool_ e1_, real_ v1_])\
   \ in Substitute [(v_1, (Rhs o hd) s_1)] h_)"
   )
,  
 prep_met 
 (("DiffApp.thy","max_on_interval_by_calculus"):metID,
   [("#Given" ,["functionTerm t_","boundVariable v_","interval itv_",
		"errorBound err_"]),
    ("#Find"  ,["maxArgument v_0_"])
    ],
   {rew_ord'="tless_true",rls'=eval_rls,calc=[],srls = e_rls,prls=e_rls,
    asm_rls=[],asm_thm=[]},
   "Undef"
   )
,
 prep_met 
 (("DiffApp.thy","find_values"):metID,
   [],
   {rew_ord'="tless_true",rls'=eval_rls,calc=[],srls = e_rls,prls=e_rls,
    asm_rls=[],asm_thm=[]},
   "Undef")
]);


