structure Test =
struct

local
 val parse_ast_translation = [];
 val parse_translation = [];
 val print_translation = [];
 val typed_print_translation = [];
 val print_ast_translation = [];
 val token_translation = [];
in



val thy = IsarThy.begin_theory "Test" ["Atools", "Rational", "Root", "Poly"] []

|> Theory.add_trfuns
(parse_ast_translation, parse_translation, print_translation, print_ast_translation)
|> Theory.add_trfunsT typed_print_translation
|> Theory.add_tokentrfuns token_translation

|> Theory.add_consts
[("Expand'_binomtest", "['y, \
                  \ 'y] => 'y", Mixfix ("((Script Expand'_binomtest (_ =))// \
                 \ (_))", [], 9)),
 ("Solve'_univar'_err", "[bool,real,bool, \
                  \ bool list] => bool list", Mixfix ("((Script Solve'_univar'_err (_ _ _ =))// \
                 \ (_))", [], 9)),
 ("Solve'_linear", "[bool,real, \
                  \ bool list] => bool list", Mixfix ("((Script Solve'_linear (_ _ =))// \
                 \ (_))", [], 9)),
 ("is'_root'_free", "'a  => bool ", Mixfix ("is'_root'_free _", [], 10)),
 ("contains'_root", "'a  => bool ", Mixfix ("contains'_root _", [], 10)),
 ("Solve'_root'_equation", "[bool,real,bool, \
                  \ bool list] => bool list", Mixfix ("((Script Solve'_root'_equation (_ _ _ =))// \
                 \ (_))", [], 9)),
 ("Solve'_plain'_square", "[bool,real, \
                  \ bool list] => bool list", Mixfix ("((Script Solve'_plain'_square (_ _ =))// \
                 \ (_))", [], 9)),
 ("Norm'_univar'_equation", "[bool,real, \
                  \ bool] => bool", Mixfix ("((Script Norm'_univar'_equation (_ _ =))// \
                 \ (_))", [], 9)),
 ("STest'_simplify", "[real, \
                  \ real] => real", Mixfix ("((Script STest'_simplify (_ =))// \
                 \ (_))", [], 9))]

|> (#1 oo (PureThy.add_axioms o map Thm.no_attributes))
[("radd_mult_distrib2", "(k::real) * (m + n) = k * m + k * n"),
 ("rdistr_right_assoc", "(k::real) + l * n + m * n = k + (l + m) * n"),
 ("rdistr_right_assoc_p", "l * n + (m * n + (k::real)) = (l + m) * n + k"),
 ("rdistr_div_right", "((k::real) + l) / n = k / n + l / n"),
 ("rcollect_right", "[| l is_const; m is_const |] ==> (l::real)*n + m*n = (l + m) * n"),
 ("rcollect_one_left", "m is_const ==> (n::real) + m * n = (1 + m) * n"),
 ("rcollect_one_left_assoc", "m is_const ==> (k::real) + n + m * n = k + (1 + m) * n"),
 ("rcollect_one_left_assoc_p", "m is_const ==> n + (m * n + (k::real)) = (1 + m) * n + k"),
 ("rtwo_of_the_same", "a + a = 2 * a"),
 ("rtwo_of_the_same_assoc", "(x + a) + a = x + 2 * a"),
 ("rtwo_of_the_same_assoc_p", "a + (a + x) = 2 * a + x"),
 ("rcancel_den", "not(a=0) ==> a * (b / a) = b"),
 ("rcancel_const", "[| a is_const; b is_const |] ==> a*(x/b) = a/b*x"),
 ("rshift_nominator", "(a::real) * b / c = a / c * b"),
 ("exp_pow", "(a ^^^ b) ^^^ c = a ^^^ (b * c)"),
 ("rsqare", "(a::real) * a = a ^^^ 2"),
 ("power_1", "(a::real) ^^^ 1 = a"),
 ("rbinom_power_2", "((a::real) + b)^^^ 2 = a^^^ 2 + 2*a*b + b^^^ 2"),
 ("rmult_1", "1 * k = (k::real)"),
 ("rmult_1_right", "k * 1 = (k::real)"),
 ("rmult_0", "0 * k = (0::real)"),
 ("rmult_0_right", "k * 0 = (0::real)"),
 ("radd_0", "0 + k = (k::real)"),
 ("radd_0_right", "k + 0 = (k::real)"),
 ("radd_real_const_eq", "[| a is_const; c is_const; d is_const |] ==> a/d + c/d = (a+c)/(d::real)"),
 ("radd_real_const", "[| a is_const; b is_const; c is_const; d is_const |] ==> a/b + c/d = (a*d + b*c)/(b*(d::real))"),
 ("radd_commute", "(m::real) + (n::real) = n + m"),
 ("radd_left_commute", "(x::real) + (y + z) = y + (x + z)"),
 ("radd_assoc", "(m::real) + n + k = m + (n + k)"),
 ("rmult_commute", "(m::real) * n = n * m"),
 ("rmult_left_commute", "(x::real) * (y * z) = y * (x * z)"),
 ("rmult_assoc", "(m::real) * n * k = m * (n * k)"),
 ("risolate_bdv_add", "((k::real) + bdv = m) = (bdv = m + (-1)*k)"),
 ("risolate_bdv_mult_add", "((k::real) + n*bdv = m) = (n*bdv = m + (-1)*k)"),
 ("risolate_bdv_mult", "((n::real) * bdv = m) = (bdv = m / n)"),
 ("rnorm_equation_add", "~(b =!= 0) ==> (a = b) = (a + (-1)*b = 0)"),
 ("root_ge0", "0 <= a ==> 0 <= sqrt a"),
 ("root_add_ge0", "[| 0 <= a; 0 <= b |] ==> (0 <= sqrt a + sqrt b) = True"),
 ("root_ge0_1", "[| 0<=a; 0<=b; 0<=c |] ==> (0 <= a * sqrt b + sqrt c) = True"),
 ("root_ge0_2", "[| 0<=a; 0<=b; 0<=c |] ==> (0 <= sqrt a + b * sqrt c) = True"),
 ("rroot_square_inv", "(sqrt a)^^^ 2 = a"),
 ("rroot_times_root", "sqrt a * sqrt b = sqrt(a*b)"),
 ("rroot_times_root_assoc", "(a * sqrt b) * sqrt c = a * sqrt(b*c)"),
 ("rroot_times_root_assoc_p", "sqrt b * (sqrt c * a)= sqrt(b*c) * a"),
 ("square_equation_left", "[| 0 <= a; 0 <= b |] ==> (((sqrt a)=b)=(a=(b^^^ 2)))"),
 ("square_equation_right", "[| 0 <= a; 0 <= b |] ==> ((a=(sqrt b))=((a^^^ 2)=b))"),
 ("square_equation", "[| 0 <= a; 0 <= b |] ==> ((a=b)=((a^^^ 2)=b^^^ 2))"),
 ("risolate_root_add", "(a+  sqrt c = d) = (  sqrt c = d + (-1)*a)"),
 ("risolate_root_mult", "(a+b*sqrt c = d) = (b*sqrt c = d + (-1)*a)"),
 ("risolate_root_div", "(a * sqrt c = d) = (  sqrt c = d / a)"),
 ("mult_square", "(a*bdv^^^2 = b) = (bdv^^^2 = b / a)"),
 ("constant_square", "(a + bdv^^^2 = b) = (bdv^^^2 = b + -1*a)"),
 ("constant_mult_square", "(a + b*bdv^^^2 = c) = (b*bdv^^^2 = c + -1*a)"),
 ("square_equality", "0 <= a ==> (x^^^2 = a) = ((x=sqrt a) | (x=-1*sqrt a))"),
 ("square_equality_0", "(x^^^2 = 0) = (x = 0)"),
 ("rroot_to_lhs", "is_root_free a ==> (a = sqrt b) = (a + (-1)*sqrt b = 0)"),
 ("rroot_to_lhs_mult", "is_root_free a ==> (a = c*sqrt b) = (a + (-1)*c*sqrt b = 0)"),
 ("rroot_to_lhs_add_mult", "is_root_free a ==> (a = d+c*sqrt b) = (a + (-1)*c*sqrt b = d)")]

|> IsarThy.end_theory;

val _ = context thy;


val radd_mult_distrib2 = PureThy.get_thm thy "radd_mult_distrib2";
val rdistr_right_assoc = PureThy.get_thm thy "rdistr_right_assoc";
val rdistr_right_assoc_p = PureThy.get_thm thy "rdistr_right_assoc_p";
val rdistr_div_right = PureThy.get_thm thy "rdistr_div_right";
val rcollect_right = PureThy.get_thm thy "rcollect_right";
val rcollect_one_left = PureThy.get_thm thy "rcollect_one_left";
val rcollect_one_left_assoc = PureThy.get_thm thy "rcollect_one_left_assoc";
val rcollect_one_left_assoc_p = PureThy.get_thm thy "rcollect_one_left_assoc_p";
val rtwo_of_the_same = PureThy.get_thm thy "rtwo_of_the_same";
val rtwo_of_the_same_assoc = PureThy.get_thm thy "rtwo_of_the_same_assoc";
val rtwo_of_the_same_assoc_p = PureThy.get_thm thy "rtwo_of_the_same_assoc_p";
val rcancel_den = PureThy.get_thm thy "rcancel_den";
val rcancel_const = PureThy.get_thm thy "rcancel_const";
val rshift_nominator = PureThy.get_thm thy "rshift_nominator";
val exp_pow = PureThy.get_thm thy "exp_pow";
val rsqare = PureThy.get_thm thy "rsqare";
val power_1 = PureThy.get_thm thy "power_1";
val rbinom_power_2 = PureThy.get_thm thy "rbinom_power_2";
val rmult_1 = PureThy.get_thm thy "rmult_1";
val rmult_1_right = PureThy.get_thm thy "rmult_1_right";
val rmult_0 = PureThy.get_thm thy "rmult_0";
val rmult_0_right = PureThy.get_thm thy "rmult_0_right";
val radd_0 = PureThy.get_thm thy "radd_0";
val radd_0_right = PureThy.get_thm thy "radd_0_right";
val radd_real_const_eq = PureThy.get_thm thy "radd_real_const_eq";
val radd_real_const = PureThy.get_thm thy "radd_real_const";
val radd_commute = PureThy.get_thm thy "radd_commute";
val radd_left_commute = PureThy.get_thm thy "radd_left_commute";
val radd_assoc = PureThy.get_thm thy "radd_assoc";
val rmult_commute = PureThy.get_thm thy "rmult_commute";
val rmult_left_commute = PureThy.get_thm thy "rmult_left_commute";
val rmult_assoc = PureThy.get_thm thy "rmult_assoc";
val risolate_bdv_add = PureThy.get_thm thy "risolate_bdv_add";
val risolate_bdv_mult_add = PureThy.get_thm thy "risolate_bdv_mult_add";
val risolate_bdv_mult = PureThy.get_thm thy "risolate_bdv_mult";
val rnorm_equation_add = PureThy.get_thm thy "rnorm_equation_add";
val root_ge0 = PureThy.get_thm thy "root_ge0";
val root_add_ge0 = PureThy.get_thm thy "root_add_ge0";
val root_ge0_1 = PureThy.get_thm thy "root_ge0_1";
val root_ge0_2 = PureThy.get_thm thy "root_ge0_2";
val rroot_square_inv = PureThy.get_thm thy "rroot_square_inv";
val rroot_times_root = PureThy.get_thm thy "rroot_times_root";
val rroot_times_root_assoc = PureThy.get_thm thy "rroot_times_root_assoc";
val rroot_times_root_assoc_p = PureThy.get_thm thy "rroot_times_root_assoc_p";
val square_equation_left = PureThy.get_thm thy "square_equation_left";
val square_equation_right = PureThy.get_thm thy "square_equation_right";
val square_equation = PureThy.get_thm thy "square_equation";
val risolate_root_add = PureThy.get_thm thy "risolate_root_add";
val risolate_root_mult = PureThy.get_thm thy "risolate_root_mult";
val risolate_root_div = PureThy.get_thm thy "risolate_root_div";
val mult_square = PureThy.get_thm thy "mult_square";
val constant_square = PureThy.get_thm thy "constant_square";
val constant_mult_square = PureThy.get_thm thy "constant_mult_square";
val square_equality = PureThy.get_thm thy "square_equality";
val square_equality_0 = PureThy.get_thm thy "square_equality_0";
val rroot_to_lhs = PureThy.get_thm thy "rroot_to_lhs";
val rroot_to_lhs_mult = PureThy.get_thm thy "rroot_to_lhs_mult";
val rroot_to_lhs_add_mult = PureThy.get_thm thy "rroot_to_lhs_add_mult";

end;
end;

open Test;
